/* epoll-over-aio test cases */

#include <stdio.h>
#include <sys/epoll.h>
#include <libaio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#define assert(x) if (!(x)) { \
    fprintf(stderr, "Assertion failed at line %d\n", __LINE__); \
    exit(1); }

#define printevent(x) printf("%s%s%s%s%s%s\n", \
    ((x) & EPOLLIN) ? "EPOLLIN " : "", \
    ((x) & EPOLLOUT) ? "EPOLLOUT " : "", \
    ((x) & EPOLLPRI) ? "EPOLLPRI " : "", \
    ((x) & EPOLLERR) ? "EPOLLERR " : "", \
    ((x) & EPOLLHUP) ? "EPOLLHUP " : "", \
    ((x) & EPOLLET) ? "EPOLLET " : "")

void pexit(char *msg) {
  perror(msg);
  exit(1);
}

io_context_t ctx;
int res, res2;
int epfd, pfd[2];
struct epoll_event evt, events[10];
int v = 123;
struct io_event ioe[10];

void test_vanilla_epoll() {
  /* create pipes */
  if (pipe(pfd) < 0)
    pexit("creating pipes");

  evt.data.fd = pfd[0];
  evt.events = EPOLLIN | EPOLLPRI | EPOLLERR | EPOLLHUP;
  if (epoll_ctl(epfd, EPOLL_CTL_ADD, pfd[0], &evt) < 0)
    pexit("epoll_ctl");

  res = epoll_wait(epfd, events, 10, 0);
  assert(res == 0);

  write(pfd[1], &v, sizeof(int));
  res = epoll_wait(epfd, events, 10, 0);
  assert(res == 1);
  assert(events[0].events & EPOLLIN);
  printevent(events[0].events);

  /* test closing the pipe */
  close(pfd[1]);
  res = epoll_wait(epfd, events, 10, 0);
  assert(res == 1);
  assert(events[0].events & EPOLLIN && events[0].events & EPOLLHUP);
  printevent(events[0].events);

  /* read the data out */
  res = read(pfd[0], &res2, sizeof(int));
  assert (res == sizeof(int));
  assert (res2 == v);

  res = epoll_ctl(epfd, EPOLL_CTL_DEL, pfd[0], &evt);
  assert (res == 0);

  close(pfd[0]);

  /* now there should be no event */
  res = epoll_wait(epfd, events, 10, 0);
  assert(res == 0);
}

void test_epio() {
  struct timespec zerotime = {.tv_sec = 0, .tv_nsec = 0};

  /* test binding a wrong epol fd */
  res = io_set_epollfd(ctx, 12345);
  assert(errno == EINVAL);

  /* bind the aio context and epoll fd */
  res = io_set_epollfd(ctx, epfd);
  if (res < 0) {
    pexit("io_set_epollfd");
  }

  /* create pipes */
  if (pipe(pfd) < 0)
    pexit("creating pipes");

  /* declare interest */
  evt.data.fd = pfd[0];
  evt.events = EPOLLIN | EPOLLPRI | EPOLLERR | EPOLLHUP;
  if (epoll_ctl(epfd, EPOLL_CTL_ADD, pfd[0], &evt) < 0)
    pexit("epoll_ctl");

  /* no data to read */
  res = io_getevents(ctx, 1, 10, ioe, &zerotime);
  assert(res == 0);

  /* write some data and get event */
  write(pfd[1], &v, sizeof(int));
  res = io_getevents(ctx, 1, 10, ioe, &zerotime);
  assert(res == 1);
  assert(ioe[0].obj == IOCB_POLL);
  assert(ioe[0].res2 == pfd[0]);
  assert(ioe[0].res & EPOLLIN);
  printevent(ioe[0].res);

  /* read the data out */
  res = read(pfd[0], &res2, sizeof(int));
  assert (res == sizeof(int));
  assert (res2 == v);

  /* now there should be no event */
  res = io_getevents(ctx, 1, 10, ioe, &zerotime);
  assert(res == 0);

  /* test event coalescing */
  v = 54321;
  write(pfd[1], &v, sizeof(int)); /* generate an event */
  close(pfd[1]);		/* update the event with EPOLLHUP */
  res = io_getevents(ctx, 1, 10, ioe, &zerotime);
  assert(res == 1);
  assert(ioe[0].res & EPOLLIN && ioe[0].res & EPOLLHUP);
  printevent(ioe[0].res);

  res = epoll_ctl(epfd, EPOLL_CTL_DEL, pfd[0], &evt);
  assert (res == 0);

  /* now there should be no event */
  res = io_getevents(ctx, 1, 10, ioe, &zerotime);
  assert(res == 0);
}

int main(int argc, char *argv[]) {
  
  /* init aio */
  res = io_setup(1024, &ctx);
  if (res != 0) {
    pexit("init aio");
  }

  /* init epoll */
  epfd = epoll_create(100);
  if (epfd < 0)
    pexit("init epoll");

  /* test vanilla epoll */
  test_vanilla_epoll();

  /* test epoll-over-aio */
  test_epio();

  close(epfd);
  io_destroy(ctx);

  printf("All tests successful\n");
  return 0;
}
