/*
 * pagefault.c
 *
 * Usage example of FLEX's user-level page-fault notification interface
 * Count and time page faults caused by accessing an mmap'ed file
 */

#include <stdio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <libaio.h>
#include <errno.h>

/* Number of reads to perform */
#define N 1000

/* for REG_EAX ... */
#define __USE_GNU 1
#include <sys/ucontext.h>

io_context_t ioctx;
int count = 0;
long long us = 0;
    
void pagefault_handler(int sig, struct siginfo *info, void *_ctx) {
    ucontext_t *ctx = (ucontext_t *)_ctx;
    struct io_event events[1];
    struct timeval t1, t2;
    (void)sig;
    
    gettimeofday (&t1, 0);
    if (count == 0) {  /* only print stuff at first page fault */
        printf("Page fault handler called!\n");
        /* We can get faulting address and all register values this way */
        printf("Faulting address: %x, ESP: %x, EBP: %x\n", (unsigned)info->si_addr, 
               ctx->uc_mcontext.gregs[REG_ESP], ctx->uc_mcontext.gregs[REG_EBP]);
    }

    /* Now wait for arrival of completion event */
    if (io_getevents(ioctx, 1, 1, events, 0) < 0) {
        perror("Cannot get completion event");
    } else {
        if (events[0].data != (unsigned long)info->si_addr 
            || events[0].obj != IOCB_PAGEFAULT) {
            perror ("Wrong event format");
        } else if (count == 0) {
            printf("Got fault completion event\n");
        }
    }
    gettimeofday(&t2, 0);
    count ++;
    us += t2.tv_usec - t1.tv_usec + 1000000*(t2.tv_sec - t1.tv_sec);
    /* printf("Got completion event.\n"); */

    /* Now the main program can continue */
}

int main(int argc, char **argv) {
    char * p;
    int fd, i, s;
    struct stat st;
    int off;
    struct sigaction sa;
    int res;

    if (argc < 2) {
        printf("Pass filename as argument\n");
        exit(0);
    }
    
    /* Set up signal handler to receive page fault notification */
    sa.sa_sigaction = pagefault_handler;
    sa.sa_flags = SA_SIGINFO;

    if (sigaction(SIGPF, &sa, 0) < 0) {
        perror ("Error setting up signal handler");
        exit(-1);
    }

    /* Set up an io context and specify it to receive page faults events */
    if (io_setup(1024, &ioctx) < 0) {
        perror("Error setting up io context");
        exit (-1);
    }
    if ((res=io_modify(ioctx, IO_SET_PAGEFAULT_CONTEXT, (void *)1)) < 0) {
        errno = -res;
        perror("Cannot set pagefault context");
        exit(-1);
    }

    /* Open and mmap an file */
    fd = open(argv[1], O_RDONLY);
    if (fd < 0) {
        perror("Error opening file");
        exit(-1);
    }
    fstat(fd, &st);
    p = mmap(0, st.st_size, PROT_READ, MAP_SHARED, fd, 0);

    if (p == MAP_FAILED) {
        perror ("Error mmapping");
        exit(-1);
    }

    /* Randomly hit some part of the file to cause page faults */
    srand(time(0));
    for (i = 0; i < N; i++) {
        off = rand() % (st.st_size);
        s += p[off];
    }

    printf("%d page faults out of %d accesses, %lld us in page faults, average %g per page fault\n", 
           count, N, us, (double)(us)/count);

    close(fd);

    io_modify(ioctx, IO_SET_PAGEFAULT_CONTEXT, 0);
    return 0;
}
