/*
 * pagefault_co.c
 *
 * Use FLEX's page fault handling interface to do some other work when a page fault happens
 * This example is limited because work is done inside a signal handler, which limits the
 * kind of syscall's the work thread can issue.
 */

#include <stdio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>
#include <signal.h>
#include <libaio.h>
#include <unistd.h>

#include "coro.h"

/* Number of reads to perform */
#define N 100

io_context_t ioctx;
    
struct coroutine *worker_co;
#define STACK_SIZE (64*1024)
char worker_stack[STACK_SIZE];

void pagefault_handler(int sig, struct siginfo *info, void *_ctx) {
    (void)sig;
    (void)info;
    (void)_ctx;

    /* Just switch to the worker coroutine */
    co_call(worker_co, 0);

    /* Now the main program can continue */
}

/* Fake worker thread to do "work" when main thread is blocked */
void worker() {
    struct io_event events[1];
    struct timespec wait;

    wait.tv_sec = 0;
    wait.tv_nsec = 0;
    
    /* Keep polling for completion of the page fault */
    while (1) {
        printf(".");
        usleep(1000);       /* 1 ms */
        if (io_getevents(ioctx, 1, 1, events, &wait) == 1) {
            printf("\n");
            co_call(co_main, 0);
        }
    }
}


int main(int argc, char **argv) {
    char * p;
    int fd, i, s;
    struct stat st;
    int off;
    struct sigaction sa;

    if (argc < 2) {
        printf("Pass filename as argument\n");
        exit(0);
    }
    
    /* Set up worker coroutine */
    worker_co = co_create (worker, worker_stack, STACK_SIZE);
    if (!worker_co) {
        perror("Error creating worker coroutine");
        exit(-1);
    }

    /* Set up signal handler to receive page fault notification */
    sa.sa_sigaction = pagefault_handler;
    sa.sa_flags = SA_SIGINFO;

    if (sigaction(SIGPF, &sa, 0) < 0) {
        perror ("Error setting up signal handler");
        exit(-1);
    }

    /* Set up an io context and specify it to receive page faults events */
    if (io_setup(1024, &ioctx) < 0) {
        perror("Error setting up io context");
        exit (-1);
    }
    if (io_modify(ioctx, IO_SET_PAGEFAULT_CONTEXT, (void *)1) < 0) {
        perror("Cannot set pagefault context");
        exit(-1);
    }

    /* Open and mmap an file */
    fd = open(argv[1], O_RDONLY);
    if (fd < 0) {
        perror("Error opening file");
        exit(-1);
    }
    fstat(fd, &st);
    p = mmap(0, st.st_size, PROT_READ, MAP_SHARED, fd, 0);

    if (p == MAP_FAILED) {
        perror ("Error mmapping");
        exit(-1);
    }

    /* Randomly hit some part of the file to cause page faults */
    srand(time(0));
    for (i = 0; i < N; i++) {
        off = rand() % (st.st_size);
        s += p[off];
    }

    printf("Done\n");

    close(fd);
    return 0;
}
