/*
 * pagefault_coro3.c
 *
 * Use FLEX's page fault handling interface to do extra work during page faults.
 * See if switching coroutine inside the signal handler is working for multiple
 * workers.
 */

#include <stdio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>
#include <signal.h>
#include <libaio.h>
#include "coro.h"

#define VERBOSE 0
#define QUIET 1

/* Number of reads to  */
int N = 100;
#define MAX_WORKER_NR 1000
#define MAX_N 10000

io_context_t ioctx;
struct coroutine *co_scheduler;
struct coroutine *co_workers[MAX_WORKER_NR];
int nr_worker;
int rands[MAX_N];
int results[MAX_WORKER_NR];
char *data;     // all data that's read

#define STACK_SIZE (64*1024)
#define STATE_RUNNABLE 0
#define STATE_BLOCKED 1
#define STATE_FINISHED 2

struct worker_context {
    /* scheduler stuff */
    int state;
    unsigned long faulting_address; /* index to match against the completion event */

    /* user stuff */
    int id;
    char *buffer;
    int length;
    char worker_stack[STACK_SIZE];
};

char scheduler_stack[STACK_SIZE];

void pagefault_handler(int sig, struct siginfo *info, void *_ctx) {
    struct worker_context *wctx;
    (void)sig;
    (void)info;

//    printf("From pagefault_handler()\n");

    /* Store faulting address in worker context */
    wctx = (struct worker_context *)co_current->user;
    wctx->state = STATE_BLOCKED;
    wctx->faulting_address = (unsigned long) info->si_addr;
//    printf("Faulting address %p\n", wctx->faulting_address);

    /* Switch to scheduler */
    co_call(co_scheduler, 0);

    /* This thread is now reenabled, just continue */
}

/* Worker coroutine that reads from a mmap'ed buffer */
void worker(void *_ctx) {
    struct worker_context *ctx;
    int i, off, checksum = 0;
    char *pdata;

    /* Get context from main coroutine */
    ctx = (struct worker_context *)_ctx;
    pdata = data + ctx->id * N;

    for (i = 0; i < N; i++) {
        off = rands[i] % (ctx->length);
#if VERBOSE        
        printf(".%d", ctx->id);
#endif
        checksum ^= (pdata[i] = ctx->buffer[off]);
    }

#if VERBOSE    
    printf("\nCoroutine %d done\n", ctx->id);
#endif
    results[ctx->id] = checksum;
    co_call(co_scheduler, (void *)(-1));
}

/* Fake scheduler coroutine */
void scheduler() {
    struct io_event events[1];
    struct timespec *wait;
    struct timespec zerowait;
    int i, j;

    zerowait.tv_sec = 0;
    zerowait.tv_nsec = 0;

    /* Keep polling for completion of the page fault */
    while (1) {
        int r;
        struct coroutine *next;
        struct worker_context *wctx;
        int alive_nr;

        /* Do not blocking waiting for events when we have something to run */
        wait = &zerowait;

wait_for_events:
        /* Test whether some coroutines become runnable again */
        if (io_getevents(ioctx, 1, 1, events, wait) == 1) {
            unsigned long address = (unsigned long) events[0].data;
            int success = 0;
            for (i = 0; i < nr_worker; i++) {
                wctx = (struct worker_context *)co_workers[i]->user;
                if (wctx->state == STATE_BLOCKED && wctx->faulting_address == address) {
                    wctx->state = STATE_RUNNABLE;
                    success = 1;
#if VERBOSE
                    printf("+%d", i);
#endif
                    break;
                }
            }
            if (!success) {
                printf("Get unmatched completion event: address=%lux\n", address);
                exit(-1);
            }
        }

        /* Scheduler a thread to run */
        next = 0;
        alive_nr = 0;
        for (i = 0; i < nr_worker; i++) {
            wctx = (struct worker_context *)co_workers[i]->user;
            if (wctx->state != STATE_FINISHED)
                alive_nr ++;
            if (wctx->state == STATE_RUNNABLE) {
                next = co_workers[i];
                break;
            }
        }
        if (alive_nr == 0) {
#if VERBOSE
            for (i = 0; i < nr_worker; i++) {
                printf("checksum[%d] = %02x < ", i, results[i] & 0xff);
                for (j = 0; j < N; j++) {
                    printf("%02x ", data[i*N+j] & 0xff);
                }
                printf("\n");
            }
#else
#if !QUIET
            for (i = 0; i < nr_worker; i++) {
                printf("checksum[%d] = %02x\n", i, results[i] & 0xff);
            }
#endif
#endif
            exit(0);
        }
        if (!next) {
            /* No thread runnable, then wait  */
#if VERBOSE            
            printf("\nX");
#endif

            wait = NULL;  /* block waiting for events */
            goto wait_for_events;
        }

        r = (int)co_call(next, next->user);
        switch (r) {
        case 0:             /* Page fault happens */
#if VERBOSE            
            printf("x");
#endif
            break;
        case -1:           /* The thread finishes */
            wctx->state = STATE_FINISHED;
            break;
        }
    }
}


int main(int argc, char **argv) {
    int i;
    struct stat st;
    struct sigaction sa;

    if (argc < 3) {
        printf("pagefault_coro3 <number of reads> <files>+\n");
        exit(0);
    }

//    printf("%p\n", signal(32, SIG_IGN));

    srand(0);

    N = atoi(argv[1]);
    if (N <= 0) {
        printf("N must > 0\n");
        exit(0);
    }
    for (i = 0; i < N; i++) {
        rands[i] = rand();
    }

    /* Set up signal handler to receive page fault notification */
    sa.sa_sigaction = pagefault_handler;
    sa.sa_flags = SA_SIGINFO | SA_NOMASK;
    sigemptyset(&sa.sa_mask);

    if (sigaction(SIGPF, &sa, 0) < 0) {
        perror ("Error setting up signal handler");
        exit(-1);
    }

    /* Set up an io context and specify it to receive page faults events */
    if (io_setup(1024, &ioctx) < 0) {
        perror("Error setting up io context");
        exit (-1);
    }
    if (io_modify(ioctx, IO_SET_PAGEFAULT_CONTEXT, (void *)1) < 0) {
        perror("Cannot set pagefault context");
        exit(-1);
    }

    srand(time(0));

    /* Create all workers */
    nr_worker = argc - 2;
    if (!(data = malloc(nr_worker * N))) {
        perror("Cannot allocate memory");
        exit(-1);
    }

    for (i = 0; i < nr_worker; i++) {
        struct worker_context *worker_context;
        int fd;

        worker_context = malloc(sizeof(struct worker_context));
        if (!worker_context) {
            perror("Error allocating worker context");
            exit(-1);
        }
        memset(worker_context, 0, sizeof(struct worker_context));
        co_workers[i] = co_create(worker, worker_context->worker_stack, STACK_SIZE);
        co_workers[i]->user = worker_context;

        worker_context->id = i;

        /* Open and mmap the file */
        fd = open(argv[i+2], O_RDONLY);
        if (fd < 0) {
            printf("Error opening file %s\n", argv[i+2]);
            perror("");
            exit(-1);
        }
        if (fstat(fd, &st)) {
            printf("Cannot stat file %s\n", argv[i+2]);
            perror("");
            exit(-1);
        }
        worker_context->buffer = mmap(0, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
        worker_context->length = st.st_size;

        if (worker_context->buffer == MAP_FAILED) {
            printf ("Error mmapping %s\n", argv[i+2]);
            perror ("");
            exit(-1);
        }
    }

    /* Create the scheduler */
    co_scheduler = co_create (scheduler, &scheduler_stack, STACK_SIZE);

    /* Start the crank */
    co_call (co_scheduler, 0);

    return 0;   /* never reached */
}
